package be.iminds.ilabt.jfed.experimenter_gui.canvas;

import be.iminds.ilabt.jfed.experimenter_gui.JFedConfiguration;
import be.iminds.ilabt.jfed.experimenter_gui.canvas.impl.*;
import be.iminds.ilabt.jfed.experimenter_gui.editor.ModelRspecUtils;
import be.iminds.ilabt.jfed.experimenter_gui.editor.bo.NodeDescription;
import be.iminds.ilabt.jfed.rspec.model.GeantTestbedType;
import be.iminds.ilabt.jfed.rspec.model.javafx_impl.*;
import javafx.beans.property.ObjectProperty;
import javafx.beans.property.SimpleObjectProperty;
import javafx.collections.ListChangeListener;
import javafx.collections.MapChangeListener;
import javafx.geometry.BoundingBox;
import javafx.geometry.Bounds;
import javafx.geometry.Point2D;
import javafx.scene.control.Button;
import javafx.scene.Node;
import javafx.scene.control.ScrollPane;
import javafx.scene.input.MouseButton;
import javafx.scene.input.MouseEvent;
import javafx.scene.layout.Pane;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.Collection;
import java.util.HashMap;
import java.util.Map;
import java.util.stream.Stream;

/**
 * User: twalcari
 * Date: 11/14/13
 * Time: 12:21 PM
 */
public abstract class ExperimentCanvas extends ScrollPane {
    private static final Logger LOG = LoggerFactory.getLogger(ExperimentCanvas.class);
    private static final double CANVAS_MARGIN = 100;
    private static final double MAX_CURSOR_NODE_DISTANCE = 40;
    private static final double SQUARED_MAX_CURSOR_TARGET_DISTANCE =
            MAX_CURSOR_NODE_DISTANCE * MAX_CURSOR_NODE_DISTANCE;
    private static final double MIN_ZOOM = 0.25, MAX_ZOOM = 5, ZOOM_FACTOR = 1.5;

    protected final FXModelRspec model;
    private final Pane pCanvas;
    private final Map<FXRspecNode, RspecCanvasNode> nodes = new HashMap<>();
    private final Map<FXRspecLink, RspecCanvasLink> links = new HashMap<>();
    private final Map<FXRspecChannel, RspecChannelCanvasNode> channels = new HashMap<>();
    private final Map<FXGeantTestbedType, RspecGTSNode> geantTestbedTypes = new HashMap<>();

    private final ListChangeListener<FXRspecNode> rspecNodeListChangeListener = change -> {
        while (change.next()) {
            if (change.wasAdded()) {
                change.getAddedSubList().forEach(this::addNodeToCanvas);
            }

            if (change.wasRemoved()) {
                change.getRemoved().forEach(this::removeNodeFromCanvas);
            }
        }
    };
    private final ListChangeListener<FXRspecLink> rspecLinkListChangeListener = change -> {
        while (change.next()) {
            if (change.wasAdded()) {
                change.getAddedSubList().forEach(this::addLinkToCanvas);
            }

            if (change.wasRemoved()) {
                change.getRemoved().forEach(this::removeLinkFromCanvas);
            }
        }
    };

    private final ListChangeListener<FXRspecChannel> rspecChannelListChangeListener = change -> {
        while (change.next()) {
            if (change.wasAdded()) {
                change.getAddedSubList().forEach(this::addChannelToCanvas);
            }

            if (change.wasRemoved()) {
                change.getRemoved().forEach(this::removeChannelFromCanvas);
            }
        }
    };

    private final ListChangeListener<FXGeantTestbedType> geantTestbedTypeListChangeListener = change -> {
        while (change.next()) {
            if (change.wasAdded()) {
                change.getAddedSubList().forEach(this::addGeantTestbedTypeToCanvas);
            }

            if (change.wasRemoved()) {
                change.getRemoved().forEach(this::removeGeantTestbedTypeFromCanvas);
            }
        }
    };
    private final ObjectProperty<CanvasStyle> canvasStyle = new SimpleObjectProperty<>(CanvasStyle.EDITABLE);

    /**
     * These variables are used to perform dragging of the canvas
     */
    private Point2D dragOrigin;
    private Point2D scrollPaneStart;
    private SelectionProvider selectionProvider = new SelectionProvider();
    private double zoom = 1;

    protected final JFedConfiguration config;
    private final CanvasNodeFactory canvasNodeFactory;

    //instance initializer
    {
        this.getStylesheets().add(ExperimentCanvas.class.getResource("canvas.css").toExternalForm());

        //create content pane
        this.pCanvas = new Pane();
        this.pCanvas.getStyleClass().add("canvas");
        setContent(pCanvas);


        //Make sure that the canvas is at least as big as the viewport of the scrollpane, otherwise the background of the
        // canvas-pane will not fill the scrollpane.
        viewportBoundsProperty().addListener((observableValue, oldValue, newValue) -> {
            if (newValue.getWidth() > pCanvas.getMinWidth()) {
                pCanvas.setMinWidth(newValue.getWidth());
            }

            if (newValue.getHeight() > pCanvas.getMinHeight()) {
                pCanvas.setMinHeight(newValue.getHeight());
            }
        });

        pCanvas.setOnMousePressed(this::onCanvasMousePressed);
        pCanvas.setOnMouseDragged(this::onCanvasMouseDragged);
        pCanvas.setOnMouseReleased(this::onCanvasMouseReleased);
        pCanvas.setOnMouseMoved(this::onCanvasMouseMoved);

        //set canvas background, and make sure that it changes when canvasStyle-field changes
        pCanvas.getStyleClass().add(canvasStyle.get().getStyleName());
        canvasStyle.addListener((observable, oldValue, newValue) -> {
            LOG.trace("Changing experiment canvas style to {}", newValue);

            pCanvas.getStyleClass().remove(oldValue.getStyleName());
            pCanvas.getStyleClass().add(newValue.getStyleName());
        });
    }


    public ExperimentCanvas(FXModelRspec model, JFedConfiguration config, CanvasNodeFactory canvasNodeFactory) {
        this.model = model;
        this.config = config;
        this.canvasNodeFactory = canvasNodeFactory;

        //registerModel model into GUI
        registerModel();
    }

    /**
     * Helper method that computes the distance between two points
     */
    private static double squareDistance(double xa, double ya, double xb, double yb) {
        return ((xa - xb) * (xa - xb)) + ((ya - yb) * (ya - yb));
    }

    public void unregisterModel() {
        model.getNodes().removeListener(rspecNodeListChangeListener);
        model.getLinks().removeListener(rspecLinkListChangeListener);
    }

    /**
     * This method is called when a mouseclick is performed on a canvas node
     */
    protected void canvasItemOnMouseClicked(MouseEvent mouseEvent) {

        if (mouseEvent.getButton() == MouseButton.PRIMARY) {

            if (mouseEvent.getSource() instanceof CanvasNode ||
                    mouseEvent.getSource() instanceof RspecCanvasLink ||
                    mouseEvent.getSource() instanceof GtsToRspecCanvasLink) {
                getSelectionProvider().setSelectedObject( mouseEvent.getSource());
            } else {
                LOG.warn("unexpected item selected: " + mouseEvent.getSource());
            }

            mouseEvent.consume();
        }
    }

    /**
     * This method is called when a dragDetected-event is caught by a canvas node
     */
    protected void canvasNodeOnDragDetected(MouseEvent mouseEvent) {
        CanvasNode node = (CanvasNode) mouseEvent.getSource();
        getSelectionProvider().setSelectedCanvasNode(node);

        mouseEvent.consume();
    }

    /**
     * This method is called when a mouseDragged-event is caught by a canvas node
     */
    protected void canvasNodeOnMouseDragged(MouseEvent mouseEvent) {
    }

    /**
     * This method handles the end of a panning-operation
     */
    protected void onCanvasMouseReleased(MouseEvent mouseEvent) {
        dragOrigin = null;
        scrollPaneStart = null;
    }

    protected void onCanvasMouseMoved(MouseEvent mouseEvent) {
    }

    private void onCanvasMousePressed(MouseEvent mouseEvent) {
        dragOrigin = new Point2D(mouseEvent.getSceneX(), mouseEvent.getSceneY());
        scrollPaneStart = new Point2D(getHvalue(), getVvalue());
    }

    /**
     * Performs panning of the canvas
     */
    protected void onCanvasMouseDragged(MouseEvent mouseEvent) {
        //perform panning instead of connecting nodes
        assert (dragOrigin != null);
        double dx = dragOrigin.getX() - mouseEvent.getSceneX();
        double dy = dragOrigin.getY() - mouseEvent.getSceneY();

        setHvalue(scrollPaneStart.getX() + dx);
        setVvalue(scrollPaneStart.getY() + dy);
    }

    private void registerModel() {
        assert (nodes.isEmpty());

        for (FXRspecNode node : model.getNodes()) {
            if (node.getNodeDescriptionId() == null) {
                NodeDescription nodeDescription = config.detectJFedNodeType(node);
                if (nodeDescription != null) {
                    node.setNodeDescriptionId(nodeDescription.getId());
                }
            }

            addNodeToCanvas(node);
        }

        model.getLinks().forEach(this::addLinkToCanvas);
        model.getChannels().forEach(this::addChannelToCanvas);
        model.getGeantTestbedTypes().forEach(this::addGeantTestbedTypeToCanvas);

        //bind model and GUI
        model.getNodes().addListener(rspecNodeListChangeListener);
        model.getLinks().addListener(rspecLinkListChangeListener);
        model.getChannels().addListener(rspecChannelListChangeListener);
        model.getGeantTestbedTypes().addListener(geantTestbedTypeListChangeListener);

    }

    public Pane getCanvas() {
        return pCanvas;
    }

    /**
     * returns a node if there is one within range.
     * Returns the closest if multiple, returns NULL if no node close enough.
     */
    public NewLinkTarget findNearbyTarget(double x, double y) {

        return Stream.concat(Stream.concat(nodes.values().stream(), geantTestbedTypes.values().stream()), links.values().stream())
                .min((o1, o2) -> Double.compare(getMinimumSquareDistanceToXY(o1, x, y), getMinimumSquareDistanceToXY(o2, x, y)))
                .filter(newLinkTarget -> getMinimumSquareDistanceToXY(newLinkTarget, x, y) < SQUARED_MAX_CURSOR_TARGET_DISTANCE)
                .orElse(null);

    }

    private double getMinimumSquareDistanceToXY(NewLinkTarget nc, double x, double y) {
        double nx = nc.getCanvasBounds().getMinX();
        if (x >= nc.getCanvasBounds().getMinX() && x <= nc.getCanvasBounds().getMaxX()) nx = x;
        else if (x > nc.getCanvasBounds().getMaxX()) nx = nc.getCanvasBounds().getMaxX();

        double ny = nc.getCanvasBounds().getMinY();
        if (y >= nc.getCanvasBounds().getMinY() && y <= nc.getCanvasBounds().getMaxY()) ny = y;
        else if (y > nc.getCanvasBounds().getMaxY()) ny = nc.getCanvasBounds().getMaxY();

        return squareDistance(x, y, nx, ny);
    }

    public static Bounds calculateBoundsInCanvas(Node node) {
        Bounds localBounds = node.getBoundsInLocal();
        return new BoundingBox(
                node.getLayoutX() + localBounds.getMinX(),
                node.getLayoutY() + localBounds.getMinY(),
                localBounds.getMinZ(),
                localBounds.getWidth(),
                localBounds.getHeight(),
                localBounds.getDepth());
    }

    /**
     * Should only be called by the addNodeToCanvasAndModel or registerModel-method
     */
    private void addNodeToCanvas(FXRspecNode rspecNode) {
        RspecCanvasNode newNode = canvasNodeFactory.createRspecCanvasNode(rspecNode, this);

        nodes.put(rspecNode, newNode);

        pCanvas.getChildren().add(newNode);

        //add event listeners
        newNode.setOnMouseClicked(this::canvasItemOnMouseClicked);
        newNode.setOnDragDetected(this::canvasNodeOnDragDetected);
        newNode.setOnMouseDragged(this::canvasNodeOnMouseDragged);

    }

    private void addLinkToCanvas(FXRspecLink rspecLink) {
        RspecCanvasLink newLink = new RspecCanvasLink(rspecLink, this);
        newLink.setOnMouseClicked(this::canvasItemOnMouseClicked);

        links.put(rspecLink, newLink);

    }

    private void removeLinkFromCanvas(FXRspecLink rspecLink) {
        RspecCanvasLink rspecCanvasLink = links.get(rspecLink);
        rspecCanvasLink.remove();
        links.remove(rspecLink);
    }

    private void addChannelToCanvas(FXRspecChannel rspecChannel) {
        RspecChannelCanvasNode newChannel = canvasNodeFactory.createRspecChannelCanvasNode(rspecChannel, this);
        newChannel.setOnMouseClicked(this::canvasItemOnMouseClicked);


        newChannel.setOnMouseDragged(this::canvasNodeOnMouseDragged);

        channels.put(rspecChannel, newChannel);

    }

    private void removeChannelFromCanvas(FXRspecChannel rspecChannel) {
        RspecChannelCanvasNode rspecChannelCanvasNode = channels.get(rspecChannel);
        rspecChannelCanvasNode.remove();
        channels.remove(rspecChannelCanvasNode);
    }

    private final Map<FXGeantTestbedType, MapChangeListener<String, GeantTestbedType.RspecInterfaceBinding>>
            gtsBindingListeners = new HashMap<>();
    private final Map<GeantTestbedType.RspecInterfaceBinding, GtsToRspecCanvasLink> gtsLinks = new HashMap<>();

    private void addGeantTestbedTypeToCanvas(final FXGeantTestbedType gtt) {
        RspecGTSNode newGTSNode = canvasNodeFactory.createRspecGTSNode(gtt, this);
        newGTSNode.setOnMouseClicked(this::canvasItemOnMouseClicked);
        newGTSNode.setOnMouseDragged(this::canvasNodeOnMouseDragged);
        geantTestbedTypes.put(gtt, newGTSNode);


        MapChangeListener<String, GeantTestbedType.RspecInterfaceBinding> portBindingListener =
                change -> {
                    if (change.wasAdded() && change.getValueAdded() != null)
                        addGtsBinding(change.getKey(), gtt, change.getValueAdded());
                    if (change.wasRemoved())
                        removeGtsBinding(change.getValueRemoved());
                };

        gtt.getPortBindings().addListener(portBindingListener);

    }

    private void removeGeantTestbedTypeFromCanvas(FXGeantTestbedType gtt) {
        gtt.getPortBindings().removeListener(gtsBindingListeners.remove(gtt));
        gtt.getPortBindings().values().forEach(this::removeGtsBinding);

        RspecGTSNode gtsNode = geantTestbedTypes.get(gtt);
        gtsNode.remove();
        geantTestbedTypes.remove(gtsNode);
    }

    private void addGtsBinding(String gtsPortName, GeantTestbedType gtt, GeantTestbedType.RspecInterfaceBinding rspecInterfaceBinding) {
        FXRspecNode node = model.getNodeByClientId(rspecInterfaceBinding.getNode());

        final GtsToRspecCanvasLink gtsToRspecCanvasLink = new GtsToRspecCanvasLink(gtsPortName, gtt, node, this);
        gtsToRspecCanvasLink.setOnMouseClicked(this::canvasItemOnMouseClicked);
        gtsLinks.put(rspecInterfaceBinding, gtsToRspecCanvasLink);
    }

    private void removeGtsBinding(GeantTestbedType.RspecInterfaceBinding rspecInterfaceBinding) {
        GtsToRspecCanvasLink link = gtsLinks.remove(rspecInterfaceBinding);
        link.remove();
    }


    private void removeNodeFromCanvas(FXRspecNode rspecNode) {
        //remove from canvas
        RspecCanvasNode rspecCanvasNode = nodes.get(rspecNode);
        pCanvas.getChildren().remove(rspecCanvasNode);

        //remove event listeners
        //add event listeners
        rspecCanvasNode.setOnMouseClicked(null);
        rspecCanvasNode.setOnDragDetected(null);
        rspecCanvasNode.setOnMouseDragged(null);


        nodes.remove(rspecNode);
    }

    public SelectionProvider getSelectionProvider() {
        return selectionProvider;
    }

    public RspecCanvasNode getRspecCanvasNode(FXRspecNode node) {
        return nodes.get(node);
    }

    public RspecCanvasLink getRspecCanvasLink(FXRspecLink link) {
        return links.get(link);
    }

    public Collection<RspecCanvasLink> getRspecCanvasLinks() {
        return links.values();
    }

    public FXModelRspec getModelRspec() {
        return model;
    }

    public void arrangeNodesOnCircle() {
        arrangeNodesOnCircle(600, 200);
    }

    public void arrangeNodesOnCircle(double width, double height) {
        ModelRspecUtils.arrangeNodesOnCircle(model, width, height);
    }

    public double getZoom() {
        return zoom;
    }

    public void setZoom(double zoom) {
        if (zoom < MIN_ZOOM || zoom > MAX_ZOOM)
            throw new IllegalArgumentException("Zoom must be between " + MIN_ZOOM + " and " + MAX_ZOOM);


        double transform = zoom / this.zoom;
        //update the location of all nodes
        for (RspecCanvasNode node : nodes.values()) {
            node.setLayoutX(node.getLayoutX() * transform);
            node.setLayoutY(node.getLayoutY() * transform);
        }


        this.zoom = zoom;
    }

    public void zoomIn() {
        double newZoom = zoom * ZOOM_FACTOR;
        if (newZoom > MAX_ZOOM)
            newZoom = MAX_ZOOM;

        setZoom(newZoom);
    }

    public void zoomOut() {
        double newZoom = zoom / ZOOM_FACTOR;
        if (newZoom < MIN_ZOOM)
            newZoom = MIN_ZOOM;

        setZoom(newZoom);
    }

    public RspecGTSNode getRspecGTSNode(GeantTestbedType gtt) {
        return geantTestbedTypes.get(gtt);
    }

    public enum CanvasStyle {
        /**
         * A canvas on which nodes can be added/removed/moved around.
         * <p>
         * The default value
         */
        EDITABLE("canvas-editable"),
        /**
         * A canvas that represents a static environment. Typically for a running experiment
         */
        FIXED("canvas-fixed"),
        /**
         * A canvas that represents a static environment, but where a fatal error occured.
         */
        FAILED("canvas-failed"),
        /**
         * A canvas that represents a static environment where the resources are not available yet.
         */
        FUTURE_RESERVATION("canvas-future-reservation"),
        /**
         * A canvas that represents a static environment where the resources have expired.
         */
        EXPIRED("canvas-expired");


        private final String styleName;

        CanvasStyle(String styleName) {
            this.styleName = styleName;
        }

        public String getStyleName() {
            return styleName;
        }
    }

    public CanvasStyle getCanvasStyle() {
        return canvasStyle.get();
    }

    public ObjectProperty<CanvasStyle> canvasStyleProperty() {
        return canvasStyle;
    }

    public void setCanvasStyle(CanvasStyle canvasStyle) {
        this.canvasStyle.set(canvasStyle);
    }
}
