package be.iminds.ilabt.jfed.experimenter_gui.gts_editor;

import be.iminds.ilabt.jfed.groovy_dsl.DSLStreamException;
import be.iminds.ilabt.jfed.gts.GtsDslConstants;
import be.iminds.ilabt.jfed.gts.GtsDslWriter;
import be.iminds.ilabt.jfed.rspec.model.ModelRspecType;
import be.iminds.ilabt.jfed.rspec.model.javafx_impl.FXModelRspec;
import be.iminds.ilabt.jfed.rspec.model.javafx_impl.FXRspecInterface;
import be.iminds.ilabt.jfed.rspec.model.javafx_impl.FXRspecNode;
import be.iminds.ilabt.jfed.rspec.parser.RspecCompare;
import be.iminds.ilabt.jfed.rspec.parser.RspecParseException;
import be.iminds.ilabt.jfed.rspec.rspec_source.RequestRspecSource;
import javafx.beans.property.*;
import javafx.collections.FXCollections;
import javafx.collections.ObservableMap;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import javax.annotation.Nullable;
import javax.xml.stream.XMLStreamException;
import java.io.File;
import java.util.List;
import java.util.Set;
import java.util.stream.Collectors;

/**
 * User: twalcari
 * Date: 5/21/2015
 * Time: 11:22
 */
public class EditableGtsDsl {
    private static final Logger LOG = LoggerFactory.getLogger(EditableGtsDsl.class);

    /**
     * The file from which this Rspec has been loaded, or where it has been saved to the last time
     */
    @Nullable
    private File file = null;

    /**
     * The name of the experiment. Typically relates to the filename of the rspec
     */
    private final StringProperty name = new SimpleStringProperty();

    private final ReadOnlyObjectWrapper<FXModelRspec> model = new ReadOnlyObjectWrapper<>();

    @Nullable
    private String lastSavedRequestRspec = null;

    private final ReadOnlyStringWrapper dsl = new ReadOnlyStringWrapper();
    private final ObservableMap<FXRspecNode, FXRspecInterface> availablePorts = FXCollections.observableHashMap();

    public EditableGtsDsl(FXModelRspec model, String name, File inputFile) {
        setModel(model);
        this.name.set(name);
        this.file = inputFile;
    }

    /**
     * Creates an empty editable rspec
     */
    public EditableGtsDsl(String name) {
        setModel(new FXModelRspec("request"));
        this.name.set(name);
    }

    public File getFile() {
        return file;
    }

    public void setFile(File file) {
        assert (file != null);
        this.file = file;
    }

    public String getName() {
        return name.get();
    }

    public void setName(String name) {
        name = name.replace("-", "_");

        this.name.set(name);
    }

    public boolean verifyName(String name) {
        return name.matches("[A-Za-z0-9_]+");
    }

    public StringProperty nameProperty() {
        return name;
    }

    public FXModelRspec getModel() {
        return model.get();
    }

    public ReadOnlyObjectProperty<FXModelRspec> modelProperty() {
        return model.getReadOnlyProperty();
    }

    public void setModel(FXModelRspec model) {
        this.model.set(model);
    }

    public void save() throws DSLStreamException {
        GtsDslWriter gtsDslWriter = new GtsDslWriter(name.get(), model.get());

        dsl.set(gtsDslWriter.call());

        List<FXRspecNode> currentPorts =
                (model.get().getNodes().filtered(node -> GtsDslConstants.GTS_URN_PORT_CM.equals(node.getComponentManagerId())));

        currentPorts.stream()
                .filter(currentPort -> !availablePorts.containsKey(currentPort))
                .forEach(currentPort -> {
                    availablePorts.put(currentPort, null);
                });

        Set<FXRspecNode> removedPorts =
                availablePorts.keySet().stream().filter(availablePort -> !currentPorts.contains(availablePort)).collect(Collectors.toSet());
        removedPorts.forEach(rspecNode -> {
            FXRspecInterface iface = availablePorts.remove(rspecNode);
            if (iface != null) {
                iface.delete();
            }
        });

        lastSavedRequestRspec = new RequestRspecSource(model.get()).getRspecXmlString();
    }

    public String getDsl() {
        return dsl.get();
    }

    public ReadOnlyStringProperty dslProperty() {
        return dsl.getReadOnlyProperty();
    }

    public ObservableMap<FXRspecNode, FXRspecInterface> getAvailablePorts() {
        return availablePorts;
    }

    public boolean isDirty() {
        if (lastSavedRequestRspec == null)
            return false;

        try {

            RspecCompare compare =
                    new RspecCompare(new RequestRspecSource(model.get()).getRspecXmlString(),
                            lastSavedRequestRspec);

            return !compare.isEqual();
        } catch (XMLStreamException | RspecParseException e) {
            LOG.warn("Could not properly parse Rspecs to compare for dirty detection", e);
            return true;
        }
    }

    public boolean revert() {
        if (lastSavedRequestRspec == null)
            return false;

        model.setValue((FXModelRspec) new RequestRspecSource(lastSavedRequestRspec, ModelRspecType.FX).getModelRspec());

        return true;
    }
}
