package be.iminds.ilabt.jfed.experimenter_gui.gts_editor;

import be.iminds.ilabt.jfed.rspec.model.RspecInterface;
import be.iminds.ilabt.jfed.rspec.model.javafx_impl.*;
import javafx.geometry.Point2D;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import static be.iminds.ilabt.jfed.gts.GtsDslConstants.*;

/**
 * User: twalcari
 * Date: 5/21/2015
 * Time: 15:36
 */
public class GTSModelEditor {

    public static final String DEFAULT_IPV4_NETMASK = "255.255.255.0";

    private static final Logger LOG = LoggerFactory.getLogger(GTSModelEditor.class);
    private final EditableGtsDsl editableGtsDsl;
    private final FXModelRspec modelRspec;

    private final FXRspecFactory rspecFactory = FXRspecFactory.getInstance();

    public GTSModelEditor(EditableGtsDsl editableGtsDsl) {
        this.editableGtsDsl = editableGtsDsl;
        this.modelRspec = editableGtsDsl.getModel();
    }

    public FXModelRspec getModelRspec() {
        return modelRspec;
    }

    public FXRspecNode addHost(Point2D location) {
        FXRspecNode node = rspecFactory.createNodeWithClientId(modelRspec, nextNodeName("host"));
        node.setComponentManagerId(GTS_URN_HOST_CM);
        //node.setComponentId(GTS_URN_LOCATION_CPH);

        node.setEditorX(location.getX());
        node.setEditorY(location.getY());

        modelRspec.getNodes().add(node);

        return node;
    }


    public FXRspecNode addPort(Point2D location) {
        FXRspecNode node = rspecFactory.createNodeWithClientId(modelRspec, nextNodeName("port"));
        node.setComponentManagerId(GTS_URN_PORT_CM);

        node.setEditorX(location.getX());
        node.setEditorY(location.getY());

        modelRspec.getNodes().add(node);

        return node;
    }


    public FXRspecNode addOFX(Point2D location) {
        FXRspecNode node = rspecFactory.createNodeWithClientId(modelRspec, nextNodeName("ofx"));
        node.setComponentManagerId(GTS_URN_OFX_CM);

        node.setEditorX(location.getX());
        node.setEditorY(location.getY());

        RspecInterface controllerIface = rspecFactory.createInterface(node, GTS_CONTROLLER_ROLE);
        controllerIface.setRole(GTS_CONTROLLER_ROLE);
        controllerIface.setPublicIpv4("10.10.100.100:9966");

        RspecInterface fabricIface = rspecFactory.createInterface(node, GTS_FABRIC_ROLE);
        fabricIface.setRole(GTS_FABRIC_ROLE);

        fabricIface.getIpAddresses().add(new FXRspecInterface.IpAddress("10.10.100.1", DEFAULT_IPV4_NETMASK, RspecInterface.IpAddress.TYPE_IPV4));

        modelRspec.getNodes().add(node);

        return node;
    }

    public FXRspecNode addExternalInterface(Point2D location) {
        FXRspecNode node = rspecFactory.createNodeWithClientId(modelRspec, nextNodeName("lab"));
        node.setComponentManagerId(GTS_URN_EXTERNAL_CM);

        node.setEditorX(location.getX());
        node.setEditorY(location.getY());

        rspecFactory.createInterface(node, GTS_TUNNEL_ID);
        modelRspec.getNodes().add(node);

        return node;
    }


    private String nextNodeName(String prefix) {
        int c = 0;
        String res = prefix + c;
        while (modelRspec.getNodeByClientId(res) != null)
            res = prefix + (++c);

        return res;
    }

    public FXRspecLink addLink(FXRspecNode node1, FXRspecNode node2) {
        FXRspecLink rspecLink = rspecFactory.createLinkWithClientId(modelRspec, modelRspec.nextLinkName(), "lan");

        addNodeToLink(node1, rspecLink);
        addNodeToLink(node2, rspecLink);

        modelRspec.getLinks().add(rspecLink);

        return rspecLink;
    }

    /**
     * Creates a new FXRspecInterface to connect an FXRspecLink and an FXRspecNode.
     * Does additional configuration on this interface (or other interfaces on the link) if
     * required by the configuration-file.
     */
    private FXRspecInterface addNodeToLink(FXRspecNode node, FXRspecLink link) {
        if (GTS_URN_EXTERNAL_CM.equals(node.getComponentManagerId())) {
            if (node.getInterfaces().size() != 1)
                throw new IllegalStateException("Expected External Interface-node to have exactly one interface 'tunnel'");

            //connect to the tunnel-port of the external interface
            FXRspecInterface tunnelIface = node.getInterfaces().get(0);
            if (tunnelIface == null)
                throw new IllegalStateException("Could not retrieve interface 'tunnel' for external interface");

            tunnelIface.bindLink(link);

            return tunnelIface;
        } else {


            FXRspecInterface iface = rspecFactory.createInterface(node,
                    link,
                    modelRspec.nextIfaceName(node));
            LOG.trace("Added node {} to link {} with iface {}", node.getUniqueId(), link.getUniqueId(), iface.getClientId());

            return iface;
        }
    }


}
