package be.iminds.ilabt.jfed.experimenter_gui.slice.tasks;

import be.iminds.ilabt.jfed.experimenter_gui.slice.ExperimentController;
import be.iminds.ilabt.jfed.experimenter_gui.slice.SliceState;
import be.iminds.ilabt.jfed.experimenter_gui.ui.status.TaskStatusIndicator;
import be.iminds.ilabt.jfed.highlevel.controller.TaskExecution;
import be.iminds.ilabt.jfed.highlevel.jobs.JobExecutor;
import be.iminds.ilabt.jfed.highlevel.jobs.ReserveSliversJob;
import be.iminds.ilabt.jfed.highlevel.jobs.ReserveSliversJobFactory;
import be.iminds.ilabt.jfed.highlevel.tasks.AllocateSliverTask;
import be.iminds.ilabt.jfed.highlevel.tasks.CreateSliverReservationTask;
import be.iminds.ilabt.jfed.highlevel.tasks.HighLevelTaskFactory;
import be.iminds.ilabt.jfed.ui.javafx.dialogs.JFDialogs;
import javafx.application.Platform;
import javafx.collections.ListChangeListener;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.concurrent.ExecutionException;
import java.util.concurrent.Future;

/**
 * User: twalcari
 * Date: 2/23/2015
 * Time: 17:48
 */
public class ReserveSliversTaskGenerator extends ExperimentTaskGenerator {
    private static final Logger LOG = LoggerFactory.getLogger(ReserveSliversTaskGenerator.class);
    private final HighLevelTaskFactory hltf;
    private final ExperimentTasksFactory experimentTasksFactory;
    private final ReserveSliversJobFactory reserveSliversJobFactory;
    private final JobExecutor jobExecutor;

    public ReserveSliversTaskGenerator(ExperimentController experimentController, HighLevelTaskFactory hltf,
                                       ExperimentTasksFactory experimentTasksFactory,
                                       ReserveSliversJobFactory reserveSliversJobFactory, JobExecutor jobExecutor) {
        super(experimentController);
        this.hltf = hltf;
        this.experimentTasksFactory = experimentTasksFactory;
        this.reserveSliversJobFactory = reserveSliversJobFactory;
        this.jobExecutor = jobExecutor;
    }

    @Override
    protected TaskStatusIndicator.Status generateExperimentTasks() {
        //now we can do the reservation
        experiment.getSlice().setRequestRspec(experiment.getNewRequestRspecSource());

        assert experiment.getSlice() != null;
        assert experiment.getSlice().getExpirationDate() != null;

        ReserveSliversJob reserveSliversJob = reserveSliversJobFactory.create(experiment.getSlice(), experiment.getRequestedEndTime(), experiment.isIncludeProjectMemberSshKeys());

        reserveSliversJob.getTaskExecutions().addListener((ListChangeListener<TaskExecution>) c -> {
            while (c.next()) {
                if (c.wasAdded()) {
                    for (TaskExecution taskExecution : c.getAddedSubList()) {
                        onNewTaskExecution(taskExecution);
                    }
                }
            }
        });

        Future<Boolean> jobFuture = jobExecutor.submitJob(reserveSliversJob);

        try {
            boolean success = jobFuture.get();

            if (success) {
                experiment.setSliceState(SliceState.FUTURE_RESERVATION);
            } else {
                experiment.getFailedReservationResources().addAll(reserveSliversJob.getFailedResources());
                experiment.setSliceState(SliceState.RESERVATION_FAILED);
            }


        } catch (InterruptedException e) {
            LOG.error("Interrupted while making reservations", e);
        }    catch (ExecutionException ex) {
            LOG.error("Error while executing ReserveSliversJob", ex);
        }

        return TaskStatusIndicator.Status.SUCCESS;
    }


    @SuppressWarnings("unchecked")
    private void onNewTaskExecution(TaskExecution taskExecution) {

        if (taskExecution.getTask() instanceof CreateSliverReservationTask) {
            experimentController.submitExperimentTask(
                    new RunningCallExperimentTask(
                            String.format("Reserving resources at %s",
                                    ((TaskExecution<CreateSliverReservationTask>) taskExecution)
                                            .getTask().getAuthority().getHrn()),
                            experiment,
                            taskExecution
                    ));


        } else if (taskExecution.getTask() instanceof AllocateSliverTask) {
            experimentController.submitExperimentTask(
                    new AllocateSliverExperimentTask(taskExecution));

        }     else{
            LOG.warn("Unexpected task-execution: {}", taskExecution);
        }
    }


    private class AllocateSliverExperimentTask extends RunningCallExperimentTask<AllocateSliverTask> {
        public AllocateSliverExperimentTask(TaskExecution<AllocateSliverTask> taskExecution) {
            super(String.format("Allocating future resources at %s", taskExecution.getTask().getAuthority().getHrn()),
                    ReserveSliversTaskGenerator.this.experiment, taskExecution);
        }

        @Override
        public void onTaskFinished(TaskExecution taskExecution) {
            if (taskExecution.getState() != TaskExecution.TaskState.SUCCESS) {
                if (getTask().hasFailedDueToNotEnoughFreeResources()) {
                    Platform.runLater(() -> {
                        updateMessage("Could not map to resources (resource shortage or unsupported topology))");
                        JFDialogs.create()
                                .owner(experimentController.getParentWindow())
                                .message(String.format("Initializing resources at %s failed, because the request could not be mapped on the resources available on the testbed. " +
                                                "This is probably caused by a resource shortage, or because the requested topology is not supported",
                                        getTask().getAuthority().getName()))
                                .masthead("Initializing resources failed")
                                .showError();
                    });
                }
            }

        }
    }
}
