package be.iminds.jfed.gts_highlevel.controller;

import be.iminds.ilabt.jfed.util.JFedPasswordManager;
import be.iminds.jfed.gts_highlevel.model.*;
import com.google.common.collect.Maps;
import javafx.beans.property.ObjectProperty;
import javafx.beans.property.SimpleObjectProperty;
import javafx.collections.FXCollections;
import javafx.collections.ObservableMap;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import javax.inject.Inject;
import javax.inject.Singleton;
import java.util.Map;

/**
 * User: twalcari
 * Date: 9/24/2015
 * Time: 15:48
 */
@Singleton
public class GtsModel {

    private final JFedPasswordManager passwordManager;


    private final ObjectProperty<GtsCredentials> credentials = new SimpleObjectProperty<>();
    private final ObservableMap<String, GtsProject> projects = FXCollections.observableHashMap();
    private final ObservableMap<String, GtsInfrastructure> infrastructures = FXCollections.observableHashMap();
    private final ObservableMap<Integer, GtsReservation> reservations = FXCollections.observableHashMap();
    private final ObservableMap<String, GtsResource> resources = FXCollections.observableHashMap();
    private final ObservableMap<TypeId, GtsType> types = FXCollections.observableHashMap();

    @Inject
    public GtsModel(JFedPasswordManager passwordManager) {
        this.passwordManager = passwordManager;
    }


    public GtsCredentials getCredentials() {
        return credentials.get();
    }

    public ObjectProperty<GtsCredentials> credentialsProperty() {
        return credentials;
    }

    public void setCredentials(GtsCredentials credentials) {
        this.credentials.set(credentials);
    }

    public String getUsername() {
        return passwordManager.getLogins("gts").get(0).getUsername();
    }

    public ObservableMap<String, GtsProject> getProjects() {
        return FXCollections.unmodifiableObservableMap(projects);
    }

    void addOrUpdateProject(GtsProject project) {
        projects.put(project.getId(), project);
    }

    /**
     * Register a project as removed
     *
     * @param project
     */
    void removeProject(GtsProject project) {
        projects.remove(project.getId());
    }

    public ObservableMap<String, GtsInfrastructure> getInfrastructures() {
        return FXCollections.unmodifiableObservableMap(infrastructures);
    }

    void addOrUpdateInfrastructure(GtsInfrastructure infrastructure) {
        infrastructures.put(infrastructure.getProviderId(), infrastructure);
    }

    /**
     * Register a infrastructure as removed
     *
     * @param infrastructure
     */
    void removeInfrastructure(GtsInfrastructure infrastructure) {
        infrastructures.remove(infrastructure.getProviderId());
    }

    public ObservableMap<Integer, GtsReservation> getReservations() {
        return reservations;
    }


    void addOrUpdateReservation(GtsReservation reservation) {
        reservations.put(reservation.getId(), reservation);
    }

    /**
     * Register a reservation as removed
     *
     * @param reservation
     */
    void removeReservation(GtsReservation reservation) {
        reservations.remove(reservation.getId());
    }


    public ObservableMap<String, GtsResource> getResources() {
        return resources;
    }


    void addOrUpdateResource(GtsResource resource) {
        resources.put(resource.getId(), resource);
    }

    /**
     * Register a resource as removed
     *
     * @param resource
     */
    void removeResource(GtsResource resource) {
        resources.remove(resource.getId());
    }


    public ObservableMap<TypeId, GtsType> getTypes() {
        return FXCollections.unmodifiableObservableMap(types);
    }

    public Map<TypeId, GtsType> getTypesInProject(GtsProject project) {
        return Maps.filterEntries(types, input -> input.getKey().projectName.equals(project.getId()));
    }


    void addOrUpdateType(GtsType type, GtsProject project) {
        types.put(new TypeId(project.getId(), type.getName()), type);
    }

    /**
     * Register a type as removed
     */
    boolean removeType(GtsType type, GtsProject project) {
        return types.remove(new TypeId(project.getId(), type.getName())) != null;
    }

    /**
     * Register a type as removed
     */
    boolean removeType(String typeName, GtsProject project) {
        return types.remove(new TypeId(project.getId(), typeName)) != null;
    }

    boolean removeProject(String projectName) {
        return projects.remove(projectName) != null;
    }

    boolean removeReservation(int reservationId) {
        return reservations.remove(reservationId) != null;
    }

    @Nullable
    public GtsReservation getReservation(int reservationId) {
        return reservations.get(reservationId);
    }

    public final static class TypeId {
        private final String projectName;
        private final String typeName;

        public TypeId(@Nonnull String projectName, @Nonnull String typeName) {
            this.projectName = projectName;
            this.typeName = typeName;
        }


        @Override
        public boolean equals(Object o) {
            if (this == o) return true;
            if (o == null || getClass() != o.getClass()) return false;

            TypeId typeId = (TypeId) o;

            if (!projectName.equals(typeId.projectName)) return false;
            return typeName.equals(typeId.typeName);

        }

        @Override
        public int hashCode() {
            int result = projectName.hashCode();
            result = 31 * result + typeName.hashCode();
            return result;
        }
    }

}
