package be.iminds.jfed.gts_highlevel.jobs;

import be.iminds.ilabt.jfed.highlevel.controller.Task;
import be.iminds.ilabt.jfed.highlevel.controller.TaskExecution;
import be.iminds.ilabt.jfed.highlevel.controller.TaskThread;
import be.iminds.ilabt.jfed.rspec.model.GeantTestbedType;
import be.iminds.ilabt.jfed.rspec.model.ModelRspecType;
import be.iminds.ilabt.jfed.rspec.model.javafx_impl.FXModelRspec;
import be.iminds.ilabt.jfed.rspec.rspec_source.RequestRspecSource;
import be.iminds.jfed.gts_highlevel.GtsHighLevelException;
import be.iminds.jfed.gts_highlevel.controller.GtsModel;
import be.iminds.jfed.gts_highlevel.model.GtsProject;
import be.iminds.jfed.gts_highlevel.model.GtsReservation;
import be.iminds.jfed.gts_highlevel.tasks.*;
import be.iminds.jfed.gts_highlevel.util.PasswordUtil;
import javafx.beans.property.ObjectProperty;
import javafx.beans.property.SimpleObjectProperty;
import javafx.collections.FXCollections;
import javafx.collections.ObservableMap;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.slf4j.Marker;
import org.slf4j.MarkerFactory;

import javax.annotation.Nonnull;
import java.util.Collections;
import java.util.concurrent.CountDownLatch;

/**
 * User: twalcari
 * Date: 10/7/2015
 * Time: 10:20
 */

public class GtsStartTestbedsJob extends AbstractGtsJob<Void> {
    private static final Logger LOG = LoggerFactory.getLogger(GtsStartTestbedsJob.class);

    private static final int PASSWORD_LENGTH = 12;
    private static final int VPN_USERNAME_SUFFIX_LENGTH = 5;

    private final String experimentName;
    private final RequestRspecSource requestRspec;
    private final GtsModel model;
    private final GtsTaskFactory gtsTaskFactory;


    // private CyclicBarrier barrier;
    private CountDownLatch finalLatch;


    //results
    private ObjectProperty<GtsProject> project = new SimpleObjectProperty<>();
    private ObservableMap<GeantTestbedType, GtsReservation> reservations = FXCollections.observableHashMap();
    private String vpnUsername;
    private String vpnPassword;


    GtsStartTestbedsJob(@Nonnull String experimentName,
                        @Nonnull RequestRspecSource requestRspec,
                        TaskThread tt,
                        GtsModel model,
                        GtsTaskFactory gtsTaskFactory) {
        super("Start GTS Testbeds in " + experimentName, tt);
        this.experimentName = experimentName;
        this.requestRspec = requestRspec;
        this.model = model;
        this.gtsTaskFactory = gtsTaskFactory;
    }

    @Override
    public Void execute() throws Exception {
        FXModelRspec modelRspec = (FXModelRspec) requestRspec.getModelRspec(ModelRspecType.FX);
        assert modelRspec != null;
        assert !modelRspec.getGeantTestbedTypes().isEmpty() : "Expected one or more GeantTestbedTypes!";

        //make sure we are logged in
        if (model.getCredentials() == null) {
            TaskExecution loginTask = submitTaskAndWait(gtsTaskFactory.createLoginGtsTask());

            if (loginTask.getState() != TaskExecution.TaskState.SUCCESS) {
                throw new GtsHighLevelException("Could not login", loginTask.getException());
            }

        }

        //create a project for this slice
        vpnUsername = experimentName + "_" + PasswordUtil.generatePassword(VPN_USERNAME_SUFFIX_LENGTH);
        vpnUsername = PasswordUtil.generatePassword(PASSWORD_LENGTH);

        CreateGtsProjectTask createGtsProjectTask =
                gtsTaskFactory.createCreateGtsProjectTask(experimentName, vpnUsername, vpnPassword,
                        Collections.singletonList(model.getUsername()));

        TaskExecution createTaskExecution = submitTaskAndWait(createGtsProjectTask);

        if (createTaskExecution.getState() != TaskExecution.TaskState.SUCCESS) {
            throw new GtsHighLevelException("Could not create a GTS project for slice " + experimentName,
                    createTaskExecution.getException());
        }

        project.setValue(createGtsProjectTask.getProject());
        assert project.get() != null;

        //   barrier = new CyclicBarrier(modelRspec.getGeantTestbedTypes().size());
        this.finalLatch = new CountDownLatch(modelRspec.getGeantTestbedTypes().size());


        modelRspec.getGeantTestbedTypes().forEach(gtt -> new StartTestbedThread(gtt).start());

        finalLatch.await();

        return null;
    }

    private TaskExecution submitTaskAndWait(Task task) throws InterruptedException {
        CountDownLatch latch = new CountDownLatch(1);

        TaskExecution taskExecution = submitTask(task, (t, te, state) -> latch.countDown());

        latch.await();

        return taskExecution;
    }

    public String getVpnPassword() {
        return vpnPassword;
    }

    public String getVpnUsername() {
        return vpnUsername;
    }

    public GtsProject getProject() {
        return project.get();
    }

    public ObjectProperty<GtsProject> projectProperty() {
        return project;
    }

    public ObservableMap<GeantTestbedType, GtsReservation> getReservations() {
        return FXCollections.unmodifiableObservableMap(reservations);
    }

    private class StartTestbedThread extends Thread {
        private final GeantTestbedType geantTestbedType;
        private final Marker logMarker;


        private StartTestbedThread(GeantTestbedType geantTestbedType) {
            this.geantTestbedType = geantTestbedType;

            setName("StartTestbedThread-" + experimentName + "-" + geantTestbedType.getName());
            this.logMarker = MarkerFactory.getMarker("GTS-" + experimentName + "-" + geantTestbedType.getName());
        }

        @Override
        public void run() {
            StartGtsResourcesTask startGtsResourcesTask =
                    gtsTaskFactory.createStartGtsResourcesTask(
                            geantTestbedType.getName(), geantTestbedType.getDsl(), project.get());

            try {
                TaskExecution taskExecution = submitTaskAndWait(startGtsResourcesTask);

                if (taskExecution.getState() != TaskExecution.TaskState.SUCCESS) {
                    LOG.error(logMarker, "Could not start GTS resources", taskExecution.getException());
                    return;
                }
                assert startGtsResourcesTask.getReservation() != null;

                reservations.put(geantTestbedType, startGtsResourcesTask.getReservation());


                UpdateGtsReservationsUntilReadyTask updateGtsReservationsUntilReadyTask =
                        gtsTaskFactory.createUpdateGtsReservationsUntilReadyTask(project.get(), startGtsResourcesTask.getReservation().getId());

                scheduleTask(updateGtsReservationsUntilReadyTask, 10000);
                /* We cannot wait for the TaskExecution to finish, as this task
                    will re-schedule itself when necessary
                    (and thus generate more TaskExecutions).
                */
                while (!updateGtsReservationsUntilReadyTask.isFinished()) {
                    Thread.sleep(1000);
                }

                //now list resources

                GtsListResourcesTask gtsListResourcesTask = gtsTaskFactory.createGtsListResourcesTask();
                submitTaskAndWait(gtsListResourcesTask);


            } catch (InterruptedException e) {
                LOG.error(logMarker, "Interrupted while starting GTS Testbed {} in slice {}", geantTestbedType.getName(), experimentName, e);
            }
            finalLatch.countDown();
        }
    }
}
