package be.iminds.ilabt.jfed.highlevel.jobs;

import be.iminds.ilabt.jfed.highlevel.controller.Task;
import be.iminds.ilabt.jfed.highlevel.controller.TaskExecution;
import be.iminds.ilabt.jfed.highlevel.controller.TaskExecutionFinishedCallback;
import be.iminds.ilabt.jfed.highlevel.controller.TaskThread;
import be.iminds.ilabt.jfed.highlevel.tasks.HighLevelTaskFactory;
import javafx.collections.FXCollections;
import javafx.collections.ObservableList;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.slf4j.Marker;
import org.slf4j.MarkerFactory;

import javax.annotation.Nonnull;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.List;
import java.util.concurrent.CountDownLatch;
import java.util.stream.Collectors;

/**
 * User: twalcari
 * Date: 7/13/2015
 * Time: 16:30
 */
public abstract class AbstractJob<V> implements Job<V> {
    private static final Logger LOG = LoggerFactory.getLogger(AbstractJob.class);

    private final Marker jobMarker;

    private final String name;

    protected final HighLevelTaskFactory hltf;
    private final TaskThread tt;

    private final ObservableList<TaskExecution> taskExecutions = FXCollections.observableArrayList();

    public AbstractJob(String name, HighLevelTaskFactory hltf, TaskThread tt) {
        super();
        this.name = name;
        this.hltf = hltf;
        this.tt = tt;

        jobMarker = MarkerFactory.getMarker(name);
    }

    @Nonnull
    public Collection<Task> getAlwaysDependsOn() {
        return Collections.emptyList();
    }

    @Nonnull
    public Collection<Task> getDependsOn() {
        return Collections.emptyList();
    }

    @Override
    public final V call() throws Exception {
        //first check and execute any dependencies that we have

        List<Task> tasksToExecute = new ArrayList<>();

        tasksToExecute.addAll(getAlwaysDependsOn());
        tasksToExecute.addAll(getDependsOn().stream()
                .filter(Task::newExecutionNeeded)
                .collect(Collectors.toList()));

        if(!tasksToExecute.isEmpty()){
            CountDownLatch latch = new CountDownLatch(tasksToExecute.size());

            List<TaskExecution> taskExecutions = new ArrayList<>();

            for (Task task : tasksToExecute) {
                TaskExecution taskExecution = submitTask(task, (t, te, state) -> latch.countDown());
                taskExecutions.add(taskExecution);
            }

            latch.await();

            //check if everything was successful

            for (TaskExecution taskExecution : taskExecutions) {
                if(taskExecution.getState() != TaskExecution.TaskState.SUCCESS){
                     LOG.error(jobMarker, "Dependency {} has state {} instead of SUCCESS. Aborting this job!",
                             taskExecution.getName(), taskExecution.getState());
                    return null;
                }
            }
        }



        //now run the real task
        return execute();
    }

    public abstract V execute() throws Exception;

    protected TaskExecution submitTask(Task t, TaskExecutionFinishedCallback... callbacks) {
        TaskExecution taskExecution = tt.addTask(t, callbacks);
        taskExecutions.add(taskExecution);

        return taskExecution;
    }

    protected TaskExecution scheduleTask(Task t, long initialDelay) {
        TaskExecution taskExecution = tt.scheduleTask(t, initialDelay);
        taskExecutions.add(taskExecution);

        return taskExecution;
    }

    @Nonnull
    @Override
    public final ObservableList<TaskExecution> getTaskExecutions() {
        return FXCollections.unmodifiableObservableList(taskExecutions);
    }



}
