package be.iminds.ilabt.jfed.highlevel.jobs;

import be.iminds.ilabt.jfed.highlevel.controller.Task;
import be.iminds.ilabt.jfed.highlevel.controller.TaskExecution;
import be.iminds.ilabt.jfed.highlevel.controller.TaskThread;
import be.iminds.ilabt.jfed.highlevel.model.*;
import be.iminds.ilabt.jfed.highlevel.tasks.*;
import be.iminds.ilabt.jfed.highlevel.util.AggregateManagerWrapperFactory;
import be.iminds.ilabt.jfed.lowlevel.GeniUserProvider;
import be.iminds.ilabt.jfed.lowlevel.api.user_spec.UserSpec;
import be.iminds.ilabt.jfed.lowlevel.api_wrapper.AggregateManagerWrapper;
import be.iminds.ilabt.jfed.lowlevel.authority.AuthorityFinder;
import be.iminds.ilabt.jfed.util.GeniUrn;
import com.google.auto.factory.AutoFactory;
import com.google.auto.factory.Provided;
import javafx.application.Platform;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.slf4j.Marker;
import org.slf4j.MarkerFactory;

import java.time.Instant;
import java.time.temporal.ChronoUnit;
import java.util.HashSet;
import java.util.List;
import java.util.Set;
import java.util.concurrent.BrokenBarrierException;
import java.util.concurrent.CountDownLatch;
import java.util.concurrent.CyclicBarrier;

/**
 * This job is meant to instantly start an experiment with the given details.
 * <p>
 * We assume that the request is stored in the slice.
 * <p>
 * It will create the slivers on all component managers that are mentioned in the request Rspec.
 * <p>
 * First we try to make all the necessary reservations. Then we try to do all the necessary allocations.
 * <p>
 * If one of these actions fail, a rollback is performed.
 * <p>
 * Then we provision all the resources, wait for them to become ready, and end by doing a listResources
 * to get an up-to-date view of the resources supplied by the authority.
 * <p>
 * User: twalcari
 * Date: 7/13/2015
 * Time: 16:48
 */
@AutoFactory
public class InstantStartSliversJob extends StartJob<Void> {
    private static final Logger LOG = LoggerFactory.getLogger(InstantStartSliversJob.class);

    private final AuthorityList authorityList;
    private final AggregateManagerWrapperFactory aggregateManagerWrapperFactory;

    private List<UserSpec> userSpecs = null;

    private CyclicBarrier barrier = null;

    InstantStartSliversJob(Slice slice, Instant requestedEndTime, boolean includeProjectMemberSshKeys,
                           @Provided HighLevelTaskFactory hltf, @Provided TaskThread tt,
                           @Provided AuthorityList authorityList,
                           @Provided AggregateManagerWrapperFactory aggregateManagerWrapperFactory,
                           @Provided GeniUserProvider geniUserProvider, @Provided EasyModel easyModel) {
        super("InstantStartSlivers for " + slice.getName(), slice, requestedEndTime.truncatedTo(ChronoUnit.SECONDS),
                includeProjectMemberSshKeys, hltf, tt, geniUserProvider, easyModel);

        this.authorityList = authorityList;
        this.aggregateManagerWrapperFactory = aggregateManagerWrapperFactory;
    }

    @Override
    public Void execute() throws UnknownAuthorityException {
        //-1. Do some preparations
        userSpecs = createUserSpecList();


        //0. First gather all the authorities that we have to contact to start this
        Set<AuthorityInfo> authorityInfos = new HashSet<>();
        for (final GeniUrn componentManagerUrn : slice.getRequestRspec().getAllComponentManagerUrns()) {
            final AuthorityInfo ai =
                    authorityList.get(authorityList.finder().findByUrn(componentManagerUrn, AuthorityFinder.Purpose.CREATE_SLIVER));

            if (ai == null) {
                LOG.error("ERROR: Component Manager Urn in Request RSpec unknown: \"" + componentManagerUrn + "\".");
                throw new UnknownAuthorityException(componentManagerUrn);
            }

            authorityInfos.add(ai);
        }
        if(authorityInfos.size() ==0){
            LOG.warn("No authorities found with something to initialize!");
            return null;
        }

        barrier = new CyclicBarrier(authorityInfos.size());

        authorityInfos.forEach(ai -> new InstantStartSliverThread(ai).start());

        return null;
    }

    private class InstantStartSliverThread extends Thread {
        private final AuthorityInfo ai;
        private final Marker logMarker;

        private boolean hasResources = false;

        private InstantStartSliverThread(AuthorityInfo ai) {
            this.ai = ai;

            setName("InstantStartSliver-" + slice.getName() + "-" + ai.getName());
            this.logMarker = MarkerFactory.getMarker(slice.getName() + "-" + ai.getName());
        }

        @Override
        public void run() {
            try {
                //gather info about authority
                final boolean supportsReservation = supportsReservation(ai);
                final boolean supportsAllocateProvision = supportsAllocateProvision(ai);


                //1. Make reservation if authority supports it
                if (supportsReservation) {
                    setAuthorityInternalState(InternalState.RESERVING);

                    Instant startTime = Instant.now().truncatedTo(ChronoUnit.SECONDS);

                    CreateSliverReservationTask csrt =
                            hltf.createSliverReservation(slice, ai, startTime, requestedEndTime, userSpecs);
                    submitTaskAndWait(csrt);

                    if (!csrt.isSuccess()) {
                        setAuthorityInternalState(InternalState.FAILED);
                        barrier.reset();
                        return; //no further cleanup to do in this thread
                    }

                    hasResources = true;
                    setAuthorityInternalState(InternalState.RESERVED);
                }

                try {
                    barrier.await();
                } catch (BrokenBarrierException e) {
                    LOG.error(logMarker, "Got broken barrier after reservation-step. Cleaning-up and aborting.", e);
                    destroyResources();
                    return;
                }

                //2. Do allocation if authority doesn't support reservation
                // If separate allocation/provision is not supported, then we do a CreateSliver immediately


                if (!supportsReservation) {
                    setAuthorityInternalState(InternalState.ALLOCATING);

                    if (supportsAllocateProvision) {
                        //do allocate call
                        AllocateSliverTask ast = hltf.allocateSliver(slice, requestedEndTime, ai);
                        TaskExecution taskExecution = submitTaskAndWait(ast);

                        if (taskExecution.getState() != TaskExecution.TaskState.SUCCESS) {
                            setAuthorityInternalState(InternalState.FAILED);
                            barrier.reset();
                            return; //no further cleanup to do in this thread
                        }
                        setAuthorityInternalState(InternalState.ALLOCATED);
                    } else {
                        //do CreateSliver-call

                        CreateSliverTask cst = hltf.createSliver(slice, requestedEndTime, ai, userSpecs);
                        TaskExecution taskExecution = submitTaskAndWait(cst);

                        if (taskExecution.getState() != TaskExecution.TaskState.SUCCESS) {
                            setAuthorityInternalState(InternalState.FAILED);
                            barrier.reset();
                            return; //no further cleanup to do in this thread
                        }
                        //we don't need to do provisioning, so we can directly wait for the resources to become ready
                        setAuthorityInternalState(InternalState.PROVISIONED);
                    }
                    hasResources = true;
                }

                try {
                    barrier.await();
                } catch (BrokenBarrierException e) {
                    LOG.error(logMarker, "Got broken barrier after allocation-step. Cleaning-up and aborting.", e);
                    destroyResources();
                    return;
                }

                //3. Provision the resources if the authority supports it.
                if (supportsAllocateProvision) {
                    ProvisionSliverTask pst = hltf.provisionSliver(slice, requestedEndTime, ai, userSpecs);

                    TaskExecution taskExecution = submitTaskAndWait(pst);

                    if (taskExecution.getState() != TaskExecution.TaskState.SUCCESS) {
                        assert taskExecution.getState() == TaskExecution.TaskState.FAILED ||
                                taskExecution.getState() == TaskExecution.TaskState.CANCELLED
                                : "Unexpected state for ProvisionSliverTask-execution : " + taskExecution.getState();

                        setAuthorityInternalState(InternalState.FAILED);

                        LOG.error(logMarker, "Could not provision resources. Cleaning up and aborting.");

                        destroyResources();
                        return;
                    }
                    setAuthorityInternalState(InternalState.PROVISIONED);
                }

                //4. Wait for the resources to become ready
                setAuthorityInternalState(InternalState.WAIT_FOR_READY);

                GetStatusUntilReadyOrFailTask getStatusUntilReadyOrFailTask =
                        hltf.getStatusUntilReadyOrFail(ai.getSfaAuthority(), slice);

                scheduleTask(getStatusUntilReadyOrFailTask, 10000);
                /* We cannot wait for the TaskExecution to finish, as this task
                    will re-schedule itself when necessary
                    (and thus generate more TaskExecutions).
                */
                while (!getStatusUntilReadyOrFailTask.isFinished()) {
                    Thread.sleep(1000);
                }

                //5. Do a list-resources to verify what we got
                submitTaskAndWait(hltf.listSliceResources(slice, ai.getSfaAuthority()));

                setAuthorityInternalState(InternalState.READY);

            } catch (InterruptedException ex) {
                LOG.error("{} was interrupted.", getName());
            }
        }

        private void destroyResources() throws InterruptedException {
            if (hasResources) {
                LOG.info(logMarker, "Requesting deletion of all slivers on authority");
                submitTaskAndWait(hltf.deleteSliversAtAuthority(slice, ai.getSfaAuthority()));
            } else {
                LOG.info(logMarker, "No resources present, so we can skip the deletion-request");
            }
        }


        private void setAuthorityInternalState(InternalState internalState) {
            Platform.runLater(() -> slice.setAuthorityInternalState(ai.getSfaAuthority(), internalState));

        }

        private TaskExecution submitTaskAndWait(Task task) throws InterruptedException {
            CountDownLatch latch = new CountDownLatch(1);

            TaskExecution taskExecution = submitTask(task, (t, te, state) -> latch.countDown());

            latch.await();

            return taskExecution;
        }

    }

    private boolean supportsAllocateProvision(AuthorityInfo ai) {
        AggregateManagerWrapper amWrapper =
                aggregateManagerWrapperFactory.getAggregateManagerWrapper(ai.getSfaAuthority());

        return amWrapper.hasSeperateAllocateAndProvision();
    }
}
