package be.iminds.ilabt.jfed.highlevel.model;

import be.iminds.ilabt.jfed.lowlevel.AnyCredential;
import be.iminds.ilabt.jfed.lowlevel.SfaCredential;
import be.iminds.ilabt.jfed.util.JFedJavaFXBindings;
import javafx.collections.*;

import javax.inject.Inject;
import javax.inject.Singleton;
import java.util.HashSet;

/**
 * ParameterHistoryModel contains credentials, urns, and other info that is useful to fill in the fields of jFed Probe.
 * The EasyModel contains an instance and adds to it, and the user can also manually add to it.
 */
@Singleton
public class ParameterHistoryModel {
    private class ObservableListSetLink<T> {
        final ObservableSet<T> set;
        final ObservableList<T> list;

        T adding = null;
        T removing = null;

        public ObservableListSetLink(final ObservableSet<T> set, final ObservableList<T> list) {
            this.set = set;
            this.list = list;
            set.addListener((SetChangeListener<T>) change -> {
                if (change.wasAdded()) {
                    T a = change.getElementAdded();
                    if (adding != null) {
                        assert adding == a;
                    } else {
                        adding = a;
                        list.add(adding);
                        adding = null;
                    }
                }
                if (change.wasRemoved()) {
                    T r = change.getElementAdded();
                    if (removing != null) {
                        assert removing == r;
                    } else {
                        removing = r;
                        list.remove(removing);
                        removing = null;
                    }
                }
            });

            list.addListener((ListChangeListener<T>) change -> {
                while (change.next()) {
                    for (T a : change.getAddedSubList()) {
                        if (adding != null) {
                            assert adding == a;
                        } else {
                            adding = a;
                            set.add(adding);
                            adding = null;
                        }
                    }
                    for (T r : change.getRemoved()) {
                        if (removing != null) {
                            assert removing == r;
                        } else {
                            removing = r;
                            set.remove(removing);
                            removing = null;
                        }
                    }
                }
            });
        }
    }


    @Inject
    public ParameterHistoryModel() {
        new ObservableListSetLink<>(userCredentialInfos, userCredentialInfosList);
        new ObservableListSetLink<>(sliceCredentialInfos, sliceCredentialInfosList);
        new ObservableListSetLink<>(uncategorizedCredentialInfos, uncategorizedCredentiaInfosList);
        new ObservableListSetLink<>(sliceUrns, sliceUrnsList);
        new ObservableListSetLink<>(sliverUrns, sliverUrnsList);
        new ObservableListSetLink<>(userUrns, userUrnsList);
    }


    public void seeUrn(String urn) {
        if (urn.startsWith("urn:publicid:IDN+")) {
            if (urn.contains("+user+"))
                addUserUrn(urn);
            if (urn.contains("+slice+"))
                addSliceUrn(urn);
            if (urn.contains("+sliver+"))
                addSliverUrn(urn);
        }
    }

    public void seeUrnsInCredential(AnyCredential credentialInfo) {
        try {
            if (credentialInfo instanceof SfaCredential) {
                seeUrn(((SfaCredential) credentialInfo).getTargetUrn());
                seeUrn(((SfaCredential) credentialInfo).getOwnerUrn());
            }
        } catch (Throwable ignored) {
        } //ignore
    }


    private final ObservableSet<AnyCredential> userCredentialInfos =
            FXCollections.observableSet(new HashSet<>());
    private final ObservableList<AnyCredential> userCredentialInfosList = FXCollections.observableArrayList();

    public ObservableSet<AnyCredential> getUserCredentials() {
        return userCredentialInfos;
    }

    public ObservableList<AnyCredential> getUserCredentialsList() {
        return userCredentialInfosList;
    }

    public void addUserCredential(AnyCredential userCredentialInfo) {
        seeUrnsInCredential(userCredentialInfo);
        userCredentialInfos.add(userCredentialInfo);
    }

    private final ObservableSet<AnyCredential> sliceCredentialInfos =
            FXCollections.observableSet(new HashSet<>());
    private final ObservableList<AnyCredential> sliceCredentialInfosList = FXCollections.observableArrayList();

    public ObservableSet<AnyCredential> getSliceCredentials() {
        return sliceCredentialInfos;
    }

    public ObservableList<AnyCredential> getSliceCredentialsList() {
        return sliceCredentialInfosList;
    }

    public void addSliceCredential(AnyCredential sliceCredentialInfo) {
        seeUrnsInCredential(sliceCredentialInfo);
        sliceCredentialInfos.add(sliceCredentialInfo);
    }

    private final ObservableSet<AnyCredential> uncategorizedCredentialInfos =
            FXCollections.observableSet(new HashSet<>());
    private final ObservableList<AnyCredential> uncategorizedCredentiaInfosList = FXCollections.observableArrayList();

    public ObservableSet<AnyCredential> getUncategorizedCredentials() {
        return uncategorizedCredentialInfos;
    }

    public ObservableList<AnyCredential> getUncategorizedCredentialsList() {
        return uncategorizedCredentiaInfosList;
    }

    public void addUncategorizedCredential(AnyCredential uncategorizedCredentiaInfo) {
        seeUrnsInCredential(uncategorizedCredentiaInfo);
        uncategorizedCredentialInfos.add(uncategorizedCredentiaInfo);
    }

    /**
     * getAnyCredentialsList
     *
     * @return the concatenation of the user, slice and uncategorized credential lists
     */
    public ObservableList<AnyCredential> getAnyCredentialsList() {
        //noinspection unchecked
        return JFedJavaFXBindings.<AnyCredential>union(userCredentialInfosList, sliceCredentialInfosList, uncategorizedCredentiaInfosList);
    }


    public ObservableList<String> getAllUrnList() {
        //noinspection unchecked
        return JFedJavaFXBindings.<String>union(sliceUrnsList, sliverUrnsList, userUrnsList);
    }


    private final ObservableSet<String> sliceUrns = FXCollections.observableSet(new HashSet<>());
    private final ObservableList<String> sliceUrnsList = FXCollections.observableArrayList();

    public ObservableSet<String> getSliceUrns() {
        return sliceUrns;
    }

    public ObservableList<String> getSliceUrnsList() {
        return sliceUrnsList;
    }

    public void addSliceUrn(String urn) {
        if (urn != null)
            sliceUrns.add(urn);
    }


    private final ObservableSet<String> sliverUrns = FXCollections.observableSet(new HashSet<>());
    private final ObservableList<String> sliverUrnsList = FXCollections.observableArrayList();

    public ObservableSet<String> getSliverUrns() {
        return sliverUrns;
    }

    public ObservableList<String> getSliverUrnsList() {
        return sliverUrnsList;
    }

    public void addSliverUrn(String urn) {
        if (urn != null)
            sliverUrns.add(urn);
    }


    private final ObservableSet<String> userUrns = FXCollections.observableSet(new HashSet<>());
    private final ObservableList<String> userUrnsList = FXCollections.observableArrayList();

    public ObservableSet<String> getUserUrns() {
        return userUrns;
    }

    public ObservableList<String> getUserUrnsList() {
        return userUrnsList;
    }

    public void addUserUrn(String urn) {
        if (urn != null)
            userUrns.add(urn);
    }
}
