package be.iminds.ilabt.jfed.highlevel.tasks;

import be.iminds.ilabt.jfed.highlevel.controller.Task;
import be.iminds.ilabt.jfed.highlevel.controller.TaskExecution;
import be.iminds.ilabt.jfed.highlevel.controller.TaskThread;
import be.iminds.ilabt.jfed.highlevel.model.Slice;
import be.iminds.ilabt.jfed.highlevel.util.AggregateManagerWrapperFactory;
import be.iminds.ilabt.jfed.lowlevel.JFedException;
import be.iminds.ilabt.jfed.lowlevel.api_wrapper.AggregateManagerWrapper;
import be.iminds.ilabt.jfed.lowlevel.api_wrapper.StatusDetails;
import be.iminds.ilabt.jfed.lowlevel.authority.SfaAuthority;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.Collections;
import java.util.List;

/**
 * User: twalcari
 * Date: 6/17/2015
 * Time: 12:22
 */
public final class GetStatusUntilReadyOrFailTask extends Task {
    private static final Logger LOG = LoggerFactory.getLogger(GetStatusUntilReadyOrFailTask.class);

    private static final long PERIOD_MS = 10_000l; //10s
    private static final int MAX_FAIL_COUNT = 5;
    private static final int MAX_BUSY_COUNT = 30 * 6; //30 minutes

    //input parameters
    private final SfaAuthority auth;
    private final Slice slice;

    //results
    private boolean finished = false;
    private int failCount = 0;
    private int busyCount = 0;
    private int triesCount = 0;
    private String status = "";
    private StatusDetails statusDetails;

    //enviroment parameters
    private final AggregateManagerWrapperFactory aggregateManagerWrapperFactory;
    private final TaskThread tt;
    private final HighLevelTaskFactory highLevelTaskFactory;



    public GetStatusUntilReadyOrFailTask(SfaAuthority auth, Slice slice,
                                         AggregateManagerWrapperFactory aggregateManagerWrapperFactory, TaskThread tt,
                                         HighLevelTaskFactory highLevelTaskFactory) {
        super("Status of Sliver @ " + auth.getUrnString());
        this.auth = auth;
        this.slice = slice;
        this.aggregateManagerWrapperFactory = aggregateManagerWrapperFactory;
        this.tt = tt;
        this.highLevelTaskFactory = highLevelTaskFactory;
    }

    @Override
    public void doTask(TaskExecution taskExecution) throws JFedException, InterruptedException {
        triesCount++;
        AggregateManagerWrapper amWrapper =
                aggregateManagerWrapperFactory.getAggregateManagerWrapper(taskExecution, auth);

        boolean busy = false;
        statusDetails = null;
        try {
            statusDetails = amWrapper.status(slice.getUrn(), slice.getCredentials());
        } catch (JFedException ex) {
            if (ex.getGeniResponseCode().isBusy()) {
                LOG.warn("Got an busy-error while fetching status of sliver. Skipping...");
                busy = true;
            } else {
                throw ex;
            }
        }

        if (busy && busyCount > MAX_BUSY_COUNT) {
            LOG.warn("Waited too long for slice to get ready \"" + slice.getUrnString() + "\"." +
                    " (Waited " + busyCount + " times 10 seconds.)");
            this.status = "Waited too long for slice to get ready. Something probably went wrong. Aborting checks.";

            return;
        }

        if (!busy && statusDetails == null) {
            failCount++;

            if (failCount < MAX_FAIL_COUNT) {
                LOG.warn("Something went wrong while waiting for slice \"" + slice.getUrnString() + "\" to become ready." +
                        " Trying again in 10 seconds...");
                tt.scheduleTask(this, PERIOD_MS);
            } else {
                LOG.warn("Something went wrong while waiting for slice \"" + slice.getUrnString() + "\" to become ready." +
                        " Will stop checking status...");
                this.status = "Fetching status failed. Aborting checks!";
            }
            return;
        }

        boolean atLeastOneFailure = false;
        if (statusDetails != null) {
            for (StatusDetails.SliverStatus sliverStatus : statusDetails.getStatusBySliverUrn().values())
                if (sliverStatus == StatusDetails.SliverStatus.FAIL)
                    atLeastOneFailure = true;
            for (StatusDetails.SliverStatus sliverStatus : statusDetails.getStatusByComponentUrn().values())
                if (sliverStatus == StatusDetails.SliverStatus.FAIL)
                    atLeastOneFailure = true;
        }

        //if not fail or ready, schedule again!
        if (busy || statusDetails.getGlobalStatus().equals(StatusDetails.SliverStatus.UNKNOWN) ||
                statusDetails.getGlobalStatus().equals(StatusDetails.SliverStatus.CHANGING)) {
            if (!atLeastOneFailure || failCount < MAX_FAIL_COUNT) { //bit of a hack
                busyCount++;
                LOG.info("Status of Slice {} is \"{}\": will check again in 10 seconds.",
                        slice.getUrnString(), statusDetails);
                tt.scheduleTask(this, PERIOD_MS);
            } else {
                failCount++;
                LOG.warn("At least one sliver has FAILED in \"" + slice.getUrnString() + "\". Retried a few times, but there's no point to keep retrying." +
                        " Will stop checking status...");
                this.status = "At least one sliver has failed.";
            }
        } else {
            if (atLeastOneFailure) {
                this.status = "At least one resource has permanently failed.";
            }

            finished = true;
        }
    }

    @Override
    public List<Task> initDependsOn() {
        return Collections.singletonList(highLevelTaskFactory.getSliceCredential(slice));
    }

    public long getMsUntilNextCheck() {
        if (!futureTasks.isEmpty()) {
            return futureTasks.get(0).getTimeLeftMs();
        } else
            return 0;

    }

    public int getTriesCount() {
        return triesCount;
    }

    public boolean isFinished() {
        return finished;
    }

    public String getStatus() {
        return status;
    }

    public StatusDetails getStatusDetails() {
        return statusDetails;
    }

    public SfaAuthority getAuthority() {
        return auth;
    }
}
