package be.iminds.ilabt.jfed.highlevel.tasks;

import be.iminds.ilabt.jfed.highlevel.controller.Task;
import be.iminds.ilabt.jfed.highlevel.controller.TaskExecution;
import be.iminds.ilabt.jfed.highlevel.controller.TaskThread;
import be.iminds.ilabt.jfed.highlevel.history.StitchingReportHistory;
import be.iminds.ilabt.jfed.highlevel.model.AuthorityList;
import be.iminds.ilabt.jfed.highlevel.model.EasyModel;
import be.iminds.ilabt.jfed.highlevel.model.Slice;
import be.iminds.ilabt.jfed.highlevel.model.UserSpecHelper;
import be.iminds.ilabt.jfed.highlevel.stitcher.ParallelStitcher;
import be.iminds.ilabt.jfed.highlevel.stitcher.StitchingReportFactory;
import be.iminds.ilabt.jfed.highlevel.util.AggregateManagerWrapperFactory;
import be.iminds.ilabt.jfed.highlevel.util.SliceRegistryUtil;
import be.iminds.ilabt.jfed.lowlevel.GeniUserProvider;
import be.iminds.ilabt.jfed.lowlevel.JFedException;
import be.iminds.ilabt.jfed.lowlevel.api.user_spec.UserSpec;
import be.iminds.ilabt.jfed.lowlevel.api_wrapper.UserAndSliceApiWrapper;
import be.iminds.ilabt.jfed.lowlevel.authority.SfaAuthority;
import be.iminds.ilabt.jfed.lowlevel.connection.JFedConnectionProvider;
import be.iminds.ilabt.jfed.preferences.JFedPreferences;
import be.iminds.ilabt.jfed.util.JFedUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.File;
import java.text.SimpleDateFormat;
import java.time.Instant;
import java.util.ArrayList;
import java.util.Date;
import java.util.List;
import java.util.concurrent.TimeUnit;

/**
 * User: twalcari
 * Date: 6/18/2015
 * Time: 10:08
 */
public final class StitchTask extends Task {
    private static final Logger LOG = LoggerFactory.getLogger(StitchTask.class);

    private final Slice slice;
    private final Instant expirationDate;
    private ParallelStitcher parallelStitcher;
    private final SfaAuthority scsAuth;
    private final boolean includeProjectMemberSshKeys;

    private ParallelStitcher.GiveUpStitchingCallback giveUpStitchingCallback;

    private final List<Task> extraDependencies = new ArrayList<>();

    private final UserAndSliceApiWrapper userAndSliceApiWrapper;
    private final AggregateManagerWrapperFactory aggregateManagerWrapperFactory;
    private final SliceRegistryUtil sliceRegistryUtil;
    private final GeniUserProvider geniUserProvider;
    private final EasyModel easyModel;
    private final TaskThread tt;
    private final JFedConnectionProvider connectionProvider;
    private final AuthorityList authorityList;
    private final JFedPreferences jFedPreferences;

    private final StitchingReportFactory stitchingReportFactory;
    private final StitchingReportHistory stitchingReportHistory;

    private final HighLevelTaskFactory highLevelTaskFactory;



    public StitchTask(Slice slice, Instant expirationDate, SfaAuthority scsAuth, boolean includeProjectMemberSshKeys,
                      ParallelStitcher.GiveUpStitchingCallback giveUpStitchingCallback,
                      UserAndSliceApiWrapper userAndSliceApiWrapper,
                      AggregateManagerWrapperFactory aggregateManagerWrapperFactory,
                      SliceRegistryUtil sliceRegistryUtil, GeniUserProvider geniUserProvider,
                      EasyModel easyModel, TaskThread tt, JFedConnectionProvider connectionProvider,
                      AuthorityList authorityList, JFedPreferences jFedPreferences,
                      StitchingReportFactory stitchingReportFactory, StitchingReportHistory stitchingReportHistory,
                      HighLevelTaskFactory highLevelTaskFactory) {
        super("Create Stitched Slivers" + (scsAuth == null ? "" : " with SCS " + scsAuth.getName()));
        this.userAndSliceApiWrapper = userAndSliceApiWrapper;
        this.aggregateManagerWrapperFactory = aggregateManagerWrapperFactory;
        this.sliceRegistryUtil = sliceRegistryUtil;
        this.geniUserProvider = geniUserProvider;
        this.easyModel = easyModel;
        this.tt = tt;
        this.connectionProvider = connectionProvider;
        this.authorityList = authorityList;
        this.jFedPreferences = jFedPreferences;
        this.stitchingReportFactory = stitchingReportFactory;
        this.stitchingReportHistory = stitchingReportHistory;
        this.highLevelTaskFactory = highLevelTaskFactory;

        assert slice != null;
        assert scsAuth != null;

        this.slice = slice;
        this.scsAuth = scsAuth;
        this.expirationDate = expirationDate;
        this.giveUpStitchingCallback = giveUpStitchingCallback;
        this.includeProjectMemberSshKeys = includeProjectMemberSshKeys;
    }

    @Override
    public void doTask(TaskExecution taskExecution) throws JFedException, InterruptedException {
        List<UserSpec> userSpecs =
                UserSpecHelper.getMergedUserSpecs(geniUserProvider, easyModel, slice.getRequestRspec().getModelRspec());
        if (includeProjectMemberSshKeys)
            userSpecs =
                    UserSpecHelper.mergeUserSpecs(userSpecs, UserSpecHelper.getSliceMemberUserSpecs(slice.getUrn(), easyModel));

        parallelStitcher =
                new ParallelStitcher(tt, easyModel, geniUserProvider,
                        slice.getRequestRspec(), connectionProvider,
                        scsAuth, userSpecs, expirationDate, authorityList,
                        jFedPreferences,
                        null /* default time between status checks*/,
                        null /* default ms before giving up on status*/,
                        null /* default ms before retry createsliver after delete*/,
                        stitchingReportFactory, aggregateManagerWrapperFactory, sliceRegistryUtil, highLevelTaskFactory);
        parallelStitcher.setGiveUpStitchingCallback(this.giveUpStitchingCallback);

        stitchingReportHistory.addStitchingReport(parallelStitcher.getStitchingReport());

        assert slice.hasCredential() :
                "Must use a slice with a known credential: " + slice + "   all slices:" + easyModel.getSlices();

        if (!userAndSliceApiWrapper.hasUserCredentials())
            throw new JFedException("UserAndSliceApiWrapper did not have user Credentials.");

        parallelStitcher.setUserCredentialAndSlice(userAndSliceApiWrapper.getCachedUserCredentialsForAM(), slice);

        for (Runnable r : preStichStartCallbacks)
            r.run();

        parallelStitcher.start();

        LOG.debug("StitchTask: stitching started. Will wait up to 2 hours for stitching to finish");
        parallelStitcher.waitUntilDone(2, TimeUnit.HOURS); //wait up to 2 hours for stitching to finish

        File jFedStitchLogDir = new File(JFedUtils.getUserDataDirectory() + File.separator + "stitch-log");
        if (!jFedStitchLogDir.exists())
            jFedStitchLogDir.mkdirs();

        SimpleDateFormat dateFormat = new SimpleDateFormat("yyyy-MM-dd_hh-mm-ss");
        String logName = "stitch-" + dateFormat.format(new Date()) + ".html";
        File logFile = new File(jFedStitchLogDir, logName);
        LOG.debug("StitchTask: stitching finished! Will write report to: " + logFile.getPath());
        parallelStitcher.getStitchingReport().toHtmlFile(logFile, false);
    }

    @Override
    public List<Task> initDependsOn() {
        List<Task> res = new ArrayList<>();
        res.add(highLevelTaskFactory.getSliceCredential(slice));
        res.add(highLevelTaskFactory.getUserSSHKeys());
        if (includeProjectMemberSshKeys)
            res.add(highLevelTaskFactory.getSliceMembersSSHKeys(slice));
        res.addAll(extraDependencies);
        return res;
    }

    public void addExtraDependency(Task t) {
        extraDependencies.add(t);
    }

    public Slice getSlice() {
        return slice;
    }

    public ParallelStitcher getParallelStitcher() {
        return parallelStitcher;
    }

    private List<Runnable> preStichStartCallbacks = new ArrayList<>();

    public void addPreStichStartCallback(Runnable r) {
        preStichStartCallbacks.add(r);
    }
}
