package be.iminds.ilabt.jfed.lowlevel;

import be.iminds.ilabt.jfed.lowlevel.authority.AuthorityListModel;
import be.iminds.ilabt.jfed.lowlevel.authority.SfaAuthority;
import be.iminds.ilabt.jfed.util.GeniUrn;
import be.iminds.ilabt.jfed.util.JFedTrustStore;
import be.iminds.ilabt.jfed.util.KeyUtil;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import javax.xml.stream.XMLInputFactory;
import javax.xml.stream.XMLStreamException;
import javax.xml.stream.XMLStreamReader;
import javax.xml.transform.stream.StreamSource;
import java.io.StringReader;
import java.security.PublicKey;
import java.security.cert.X509Certificate;
import java.util.ArrayList;
import java.util.Date;
import java.util.List;

public class AbacCredential extends AnyCredential {
    private static final Logger LOG = LoggerFactory.getLogger(AbacCredential.class);

    protected String name;
    protected String credentialXml;

    protected String type;
    protected String version;

    //whether process() ran
    private boolean processed = false;
    //fields derived using process():
    private boolean speaksFor;
    private GeniUrn spokenForUrn;
    private PublicKey spokenForPubKey;
    private List<String> signerCerts;

    protected AbacCredential(String name, String credentialXml, String type, String version) {
        super(name, credentialXml, type, version);

        this.name = name;
        this.credentialXml = credentialXml;
        this.type = type;
        this.version = version;

        assert type.equalsIgnoreCase("abac") || type.equalsIgnoreCase("geni_abac") :
                "Created SfaCredential not of type sfa, but of type=\"" + type + "\" version=\"" + version + "\"";

        if (credentialXml == null) throw new RuntimeException("AbacCredential credentialXml may not be null");
    }

    //TODO
    @Override
    public String getExpires() {
        return null;
    }

    @Override
    public Date getExpiresDate() {
        return null;
    }

    public String getCredentialXml() {
        assert credentialXml != null;
        return credentialXml;
    }

    @Override
    public boolean isSpeaksFor() {
        process();
        return speaksFor;
    }

    public GeniUrn getSpokenForUrn() {
        process();
        return spokenForUrn;
    }

    public PublicKey getSpokenForPubKey() {
        process();
        return spokenForPubKey;
    }

//    public String getSignerCert() {
//        process();
//        return signerCert;
//    }

    private void process() {
        if (processed) return;
        processed = true;

        speaksFor = false;
        signerCerts = new ArrayList<String>();
        spokenForUrn = null;
        spokenForPubKey = null;

        /* speaksFor can be detected with:
                 <signed-credential xsi:noNamespaceSchemaLocation="http://www.protogeni.net/resources/credential/credential.xsd" >
                  <credential xml:id="_0">
                	<abac>
                		<head>
                            <role>speaks_for_*</role>

           signer cert is in:
              <signed-credential xsi:noNamespaceSchemaLocation="http://www.protogeni.net/resources/credential/credential.xsd">
              <signatures>
                  <Signature xmlns="http://www.w3.org/2000/09/xmldsig#">
                    <KeyInfo>
                      <X509Data>
                        <X509Certificate>
                * */

        XMLStreamReader streamReader = null;
        try {
            XMLInputFactory xif = XMLInputFactory.newFactory();
            StreamSource xml = new StreamSource(new StringReader(credentialXml));
            streamReader = xif.createXMLStreamReader(xml);

            boolean withinSignedCredential = false;

            boolean withinCredential = false;
            boolean withinAbac = false;
            boolean withinHead = false;
            boolean withinRole = false;

            boolean withinSignatures = false;
            boolean withinSignature = false;
            boolean withinKeyInfo = false;
            boolean withinX509Data = false;
            boolean withinX509Certificate = false;

            String x509certText = "";
            while (streamReader.hasNext()) {
                streamReader.next();
                switch (streamReader.getEventType()) {
                    case XMLStreamReader.START_ELEMENT: {
                        if (streamReader.getName().getLocalPart().equals("signed-credential"))
                            withinSignedCredential = true;

                        if (withinSignedCredential && streamReader.getName().getLocalPart().equals("credential"))
                            withinCredential = true;
                        if (withinCredential && streamReader.getName().getLocalPart().equals("abac"))
                            withinAbac = true;
                        if (withinAbac && streamReader.getName().getLocalPart().equals("head"))
                            withinHead = true;
                        if (withinHead && streamReader.getName().getLocalPart().equals("role"))
                            withinRole = true;

                        if (withinSignedCredential && streamReader.getName().getLocalPart().equals("signatures"))
                            withinSignatures = true;
                        if (withinSignatures && streamReader.getName().getLocalPart().equals("Signature"))
                            withinSignature = true;
                        if (withinSignature && streamReader.getName().getLocalPart().equals("KeyInfo"))
                            withinKeyInfo = true;
                        if (withinKeyInfo && streamReader.getName().getLocalPart().equals("X509Data"))
                            withinX509Data = true;
                        if (withinX509Data && streamReader.getName().getLocalPart().equals("X509Certificate")) {
                            withinX509Certificate = true;
                            x509certText = "";
                        }
                        break;
                    }
                    case XMLStreamReader.END_ELEMENT: {
                        if (withinSignedCredential && streamReader.getName().getLocalPart().equals("signed-credential"))
                            withinSignedCredential = false;

                        if (withinSignatures && streamReader.getName().getLocalPart().equals("credential"))
                            withinCredential = false;
                        if (withinCredential && streamReader.getName().getLocalPart().equals("abac"))
                            withinAbac = false;
                        if (withinAbac && streamReader.getName().getLocalPart().equals("head"))
                            withinHead = false;
                        if (withinHead && streamReader.getName().getLocalPart().equals("role"))
                            withinRole = false;

                        if (withinSignedCredential && streamReader.getName().getLocalPart().equals("signatures"))
                            withinSignatures = false;
                        if (withinSignatures && streamReader.getName().getLocalPart().equals("Signature"))
                            withinSignature = false;
                        if (withinSignature && streamReader.getName().getLocalPart().equals("KeyInfo"))
                            withinKeyInfo = false;
                        if (withinKeyInfo && streamReader.getName().getLocalPart().equals("X509Data"))
                            withinX509Data = false;
                        if (withinX509Data && streamReader.getName().getLocalPart().equals("X509Certificate")) {
                            if (!x509certText.trim().isEmpty()) {
                                if (x509certText != null) {
                                    if (!x509certText.startsWith("-----BEGIN CERTIFICATE-----"))
                                        x509certText = "-----BEGIN CERTIFICATE-----\n" + x509certText + "\n-----END CERTIFICATE-----\n";
                                    signerCerts.add(x509certText.trim());
                                }
                            }

                            withinX509Certificate = false;
                            x509certText = "";
                        }
                        break;
                    }
                    case XMLStreamReader.CHARACTERS: {
                        if (withinRole) {
                            String roleText = streamReader.getText();
                            if (roleText != null) {
                                speaksFor = roleText.trim().toLowerCase().startsWith("speaks_for_");
                            }
                        }
                        if (withinX509Certificate) {
                            x509certText += streamReader.getText();
                        }
                        break;
                    }
                    default:
                        break;
                }
            }
        } catch (XMLStreamException e) {
            LOG.error("Exception while parsing ABAC credential", e);
            speaksFor = false;
        } finally {
            if (streamReader != null)
                try {
                    streamReader.close();
                } catch (XMLStreamException e) {
                    LOG.error("Exception closing streamReader is ignored", e);
                }
        }

        if (speaksFor) {
            List<GeniUrn> subjectUrns = new ArrayList<GeniUrn>();
            List<GeniUrn> issuerUrns = new ArrayList<GeniUrn>();
            for (String signerCert : signerCerts) {
                X509Certificate cert = KeyUtil.pemToX509Certificate(signerCert);
                if (cert == null) {
                    LOG.warn("Failed to convert signerCert to X509 certificate: \"\"\"\n" + signerCert + "\n\"\"\"");
                } else {
                    subjectUrns.addAll(KeyUtil.findUrnsInCertAltNames(cert, KeyUtil.AltNamesSource.SUBJECT_ALT_NAMES, false));
                    issuerUrns.addAll(KeyUtil.findUrnsInCertAltNames(cert, KeyUtil.AltNamesSource.ISSUES_ALT_NAMES, false));
                }
            }

            //simple way to handle most certificate chains
            List<GeniUrn> possibleUserUrns = new ArrayList<GeniUrn>(subjectUrns);
            possibleUserUrns.removeAll(issuerUrns);

            if (!possibleUserUrns.isEmpty()) {
                if (possibleUserUrns.size() > 1) {
                    System.err.println("Warning: more than 1 possible user urn in speaksFor credential signer subject alt names: " + possibleUserUrns);
                    LOG.warn("Warning: more than 1 possible user urn in speaksFor credential signer subject alt naems: " + possibleUserUrns);
                }
                spokenForUrn = possibleUserUrns.get(0);

                //and finally, serach for hte publickey matching this URN
                for (String signerCert : signerCerts) {
                    X509Certificate cert = KeyUtil.pemToX509Certificate(signerCert);
                    if (cert == null) {
                        LOG.warn("Failed to convert signerCert to X509 certificate: \"\"\"\n" + signerCert + "\n\"\"\"");
                    } else {
                        spokenForPubKey = cert.getPublicKey();
                    }
                }

            }

        }
    }

    @Override
    public boolean check(AuthorityListModel authorityListModel) throws CredentialException {
        process();

        JFedTrustStore jFedTrustStore = new JFedTrustStore();

        for (String signerCert : signerCerts) {
            X509Certificate cert = KeyUtil.pemToX509Certificate(signerCert);
            List<GeniUrn> urns = KeyUtil.findUrnsInCertAltNames(cert, KeyUtil.AltNamesSource.SUBJECT_ALT_NAMES, false);
            urns.addAll(KeyUtil.findUrnsInCertAltNames(cert, KeyUtil.AltNamesSource.ISSUES_ALT_NAMES, false));
            for (GeniUrn urn : urns) {
                SfaAuthority auth = authorityListModel.getFromAnyUrn(urn, AuthorityListModel.SubAuthMatchAllowed.ALLOW_TOPLEVEL, AuthorityListModel.SubAuthMatchPreference.PREFER_EXACT_SUBAUTHORITY);
                jFedTrustStore.addAuthorityCert(auth);
            }
        }

        return check(jFedTrustStore.getTrustStore());
    }

    @Override
    public boolean equals(Object o) {
        if (this == o) return true;
        if (o == null || getClass() != o.getClass()) return false;

        AbacCredential that = (AbacCredential) o;

        if (!credentialXml.equals(that.credentialXml)) return false;
        if (!name.equals(that.name)) return false;

        return true;
    }

    @Override
    public int hashCode() {
        int result = name.hashCode();
        result = 31 * result + credentialXml.hashCode();
        return result;
    }
}
